#!/usr/bin/env bash

# uri.sh - URI processing functions.
# Copyright (c) 2022 ge <https://nixhacks.net/>
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.

parse_uri() {
    # Universal URI parser. There is fantastic shitcode, but it works.
    # This parser is mostly compatible with RFC3986:
    # https://datatracker.ietf.org/doc/html/rfc3986
    # URL-encoded passwords supported.
    #
    # Usage: parse_uri URI
    # Return variables:
    #   scheme
    #   username
    #   password
    #   hostname
    #   port
    #   path
    #   query
    #   fragment

    local uri
    local authority
    local userinfo
    local host
    local ipv6_hostname

    uri="$1"

    # * Get scheme
    scheme="${uri%%:*}"

    # * Get authority component
    # The absence of two slashes (//) after the scheme means that the path
    # component follows, not the authority.
    if [[ "$uri" =~ ${scheme}:[^//] ]]; then
        # Example (empty authority): scheme:/path
        authority=
    elif [[ "$uri" =~ ${scheme}:/// ]]; then
        # Example (empty authority): scheme:///path
        authority=
    else
        # Exapmle: scheme://authority/path

        # Fix 'scheme://authority' test failure
        # Correctly detect authority component if empty path is set
        if <<< "$uri" sed 's/\/\///g' | grep / &>/dev/null; then
            # URI contain non empty path component
            authority="$(<<< "$uri" grep -Po '(?<=//)(.[^/]+)(?=/)')" || true
        else
            # If URI does not contain path starts with "/"
            authority="$(<<< "$uri" grep -Po '(?<=//)(.[^/]+)$')" || true
        fi
    fi

    # * Get host and userinfo components if authority is set.
    if [[ "$authority" =~ @ ]]; then
        userinfo="$(<<< "$authority" cut -d '@' -f 1)" || true
        host="$(<<< "$authority" cut -d '@' -f 2)" || true
    else
        userinfo=
        host="$authority"
    fi

    # * Get fragment
    fragment="$(<<< "$uri" grep -Po '(?<=#)(.*)')" || true

    # * Get query
    query="$(<<< "$uri" grep -Po '(?<=\?)(.*)' | sed "s/#${fragment}//g")" || true

    # * Get path
    if [[ "$uri" =~ ^${scheme}:// ]]; then
        path="$(<<< "$uri" sed "s/${scheme}:\/\/${authority//\[/\\[}//g")" || true
        path="$(<<< "$path" sed "s/\?${query}//g;s/#${fragment}//g")" || true
        # Dirty hack for 'schema://host:~/path'
        if [[ "$uri" =~ :~/ ]]; then
            path="~${path}"
        fi
    else
        # For non authority component URI
        path="$(<<< "$uri" sed "s/${scheme}://g")" || true
    fi

    # * Get hostname and port
    ipv6_hostname="$(<<< "$host" grep -Po '\[(.*)\]')" || true
    if [ "$ipv6_hostname" ]; then
        hostname="$ipv6_hostname"
        port="$(<<< "$host" awk -F ']:' '{print $2}')"
    elif [[ "$host" =~ : ]]; then
        hostname="$(<<< "$host" cut -d ':' -f 1)"
        port="$(<<< "$host" cut -d ':' -f 2)"
        # Dirty hack for 'schema://host:~/path'
        if [[ "$port" == '~' ]]; then
            port=
        fi
    else
        hostname="$host"
        port=
    fi

    # * Get username and password
    if [[ "$userinfo" =~ : ]]; then
        username="$(<<< "$userinfo" cut -d ':' -f 1)"
        password="$(<<< "$userinfo" cut -d ':' -f 2)"
    else
        username="$userinfo"
        password=
    fi

    # Debug
    # echo "SCHEME=<$scheme>"
    # echo "AUTHORITY=<$authority>"
    # echo "USERINFO=<$userinfo>"
    # echo "USERNAME=<$username>"
    # echo "PASSWORD=<$password>"
    # echo "HOST=<$host>"
    # echo "HOSTNAME=<$hostname>"
    # echo "PORT=<$port>"
    # echo "PATH=<$path>"
    # echo "QUERY=<$query>"
    # echo "FRAGMENT=<$fragment>"

    # Decode URL-encoded password
    if [ "$password" ]; then
        password="$(<<< "$password" sed 's/+/ /g;s/%/\\\\x/g' | xargs echo -e)"
    fi
}
